﻿-- locals, helperfunctions
local string_gsub, string_format = string.gsub, string.format
local table_sort, table_concat, wipe = table.sort, table.concat, wipe

-- LibDataBroker
local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
local LDBIcon = ldb and LibStub("LibDBIcon-1.0",true)
local LibQTip = LibStub('LibQTip-1.0')
local Ailo = LibStub("AceAddon-3.0"):NewAddon("Ailo", "AceConsole-3.0", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Ailo", false)
local DB_VERSION = 1
local RAID_CLASS_COLORS = RAID_CLASS_COLORS
local raidorder, currentChar = {}
local function setColor(info, r, g, b, a)
    Ailo.db.profile[info[#info]] = { r = r, g = g, b = b, a = a }
end

local function getColor(info)
    return Ailo.db.profile[info[#info]].r, Ailo.db.profile[info[#info]].g, Ailo.db.profile[info[#info]].b, Ailo.db.profile[info[#info]].a
end

local defaults = {
    profile = {
        savedraid = { r=1, g=0, b=0, a=1 },
        freeraid  = { r=0, g=1, b=0, a=1 },
        hc = "hc",
        nhc = "nhc",
        showMessages = true,
        useClassColors = true,
        show5Man = false,
        instanceAbbr = {},
        minimapIcon = {
            hide = false,
            minimapPos = 220,
            radius = 80,
        },
    },
    global = {
        chars = {},
        charClass = {},
        raids = {},
        nextPurge = 0,
        version = 0,
    },
}


function Ailo:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("AiloDB", defaults)
    self:RegisterEvent("UPDATE_INSTANCE_INFO")
    self:RegisterEvent("CHAT_MSG_SYSTEM")
    LibStub("AceConfig-3.0"):RegisterOptionsTable("Ailo", self.GenerateOptions)

    currentChar = UnitName("player")

    if not self.db.global.version or DB_VERSION > self.db.global.version then
        self:Output(L["DB_VERSION_UPGRADE_PURGE"])
        Ailo:WipeDB()
        self.db.global.version = DB_VERSION 
    end

    local AiloLDB = ldb:NewDataObject("Ailo", {
        type = "data source",
        text = "Ailo",
        icon = "Interface\\Icons\\Achievement_Dungeon_UlduarRaid_Archway_01.png",
        OnClick = function(clickedframe, button)
            if button == "RightButton" then 
                InterfaceOptionsFrame_OpenToCategory(Ailo.optionsFrame) 
            else 
                Ailo:ManualPlayerUpdate() 
            end
        end,
        OnEnter = function(tt)
            local tooltip = LibQTip:Acquire("AiloTooltip", 1, "LEFT") 
            Ailo.tooltip = tooltip
            Ailo:PrepareTooltip(tooltip) 
            tooltip:SmartAnchorTo(tt)
            tooltip:Show()
        end,
        OnLeave = function(tt)
            LibQTip:Release(Ailo.tooltip)
            Ailo.tooltip = nil
        end,
    })

    LDBIcon:Register("Ailo", AiloLDB, self.db.profile.minimapIcon)
    -- Request saved raidID's for this char
    -- Will trigger UPDATE_INSTANCE_INFO when after the data is recieved
    RequestRaidInfo()
    self:SetupClasscoloredFonts()
end

local RAID_CLASS_COLORS_FONTS = {}

function Ailo:SetupClasscoloredFonts()
    local class,color
    for class,color in pairs(RAID_CLASS_COLORS) do
        RAID_CLASS_COLORS_FONTS[class] = CreateFont("ClassFont"..class)
        RAID_CLASS_COLORS_FONTS[class]:CopyFontObject(GameTooltipText)
        RAID_CLASS_COLORS_FONTS[class]:SetTextColor(color.r, color.g, color.b)
    end
end

function Ailo:GenerateOptions()
    Ailo.options = {
        name = "Ailo",
        type = 'group',
        args = {
            genconfig = {
                name = L["General Settings"],
                type = 'group',
                order = 1,
                args = {
                    savedraid = {
                        name = L["Saved raid color"],
                        desc = L["SAVED_RAID_DESC"],
                        type = 'color',
                        order = 1,
                        get  = getColor,
                        set  = setColor,
                        hasAlpha = true,
                    },
                    freeraid = {
                        name = L["Free raid color"],
                        desc = L["FREE_RAID_DESC"],
                        type = 'color',
                        order = 2,
                        get  = getColor,
                        set  = setColor,
                        hasAlpha = true,
                    },
                    minimapIcon = {
                        type = "toggle",
                        name = L["Show minimap button"],
                        desc = L["Show the Ailo minimap button"],
                        order = 3,
                        get = function(info) return not Ailo.db.profile.minimapIcon.hide end,
                        set = function(info, value)
                            Ailo.db.profile.minimapIcon.hide = not value
                            if value then LDBIcon:Show("Ailo") else LDBIcon:Hide("Ailo") end
                        end,
                    },
                    useClassColors = {
                        type = "toggle",
                        name = L["Color names by class"],
                        get = function(info) return Ailo.db.profile[info[#info]] end,
                        set = function(info, value) Ailo.db.profile[info[#info]] = value end,
                    },
                    hc = {
                        type = "input",
                        name = L["Tooltip abbreviation used for heroic raids"],
                        get = function(info) return Ailo.db.profile[info[#info]] end,
                        set = function(info, value) Ailo.db.profile[info[#info]] = value end,
                    },
                    nhc = {
                        type = "input",
                        name = L["Tooltip abbreviation used for nonheroic raids"],
                        get = function(info) return Ailo.db.profile[info[#info]] end,
                        set = function(info, value) Ailo.db.profile[info[#info]] = value end,
                    },
                    show5Man = {
                        type = "toggle",
                        name = L["Show 5-man instances"],
                        get = function(info) return Ailo.db.profile[info[#info]] end,
                        set = function(info, value) Ailo.db.profile[info[#info]] = value end,
                    },
                    showMessages = {
                        type = "toggle",
                        name = L["Chatmessages"],
                        get = function(info) return Ailo.db.profile[info[#info]] end,
                        set = function(info, value) Ailo.db.profile[info[#info]] = value end,
                    },
                    wipeDB = {
                        type = "execute",
                        name = L["Wipe Database"],
                        order = -1,
                        confirm = true,
                        func = function() Ailo:WipeDB() end,
                    },
                },
            },
            instanceAbbr = { 
                type = 'group',
                name = L["Instance Abbreviations"],
                args = {
                    header = {
                        type = "header",
                        name = L["Change the abbreviations used in the tooltip"]
                    },
                },
            },
        },
    }
    local instance, abbr
    for instance, abbr in pairs(Ailo.db.profile.instanceAbbr) do
        Ailo.options.args.instanceAbbr.args[instance] = {
            type = "input",
            name = instance,
            get = function(info) return Ailo.db.profile.instanceAbbr[info[#info]] end,
            set = function(info, value) Ailo.db.profile.instanceAbbr[info[#info]] = value end,
        }
    end
    Ailo.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(Ailo.db)
    return Ailo.options
end

function Ailo:Output(...)
    if self.db.profile.showMessages then
        Ailo:Print(...)
    end
end 

function Ailo:PrepareTooltip(tooltip)
    --[[[ Cell are just colored green/red
               ToC       VoA
            10     25   10  25
          hc nhc hc nhc nhc nhc
    Char1 [] [ ] [] [ ] [ ] [ ]
    Char2 [] [ ] [] [ ] [ ] [ ]
    Char3 [] [ ] [] [ ] [ ] [ ]
    ]]--
    local charsdb = self.db.global.chars
    local raidsdb = self.db.global.raids
    
    if next(charsdb) then
        -- At least one char is saved to some 
        tooltip:AddHeader(L["Raid"]) -- Raid
        tooltip:AddHeader(L["Size"]) -- Size
        tooltip:AddHeader(L["Diff"]) -- Heroic
        local nextPurge = self.db.global.nextPurge
        if nextPurge > 0 and time() > nextPurge then
            -- Only search 
            self:PurgeOldRaidIDs()
            self:TrimRaidTable()
            self.db.global.nextPurge = self:GetNextPurge()
        end

        local raid, size, difficulties,  difficulty, colcount
        local colcount, numdifficulties
        for raid, sizes in pairs(raidsdb) do
            colcount = 0 -- Span needed for the 'Raid' cell above the 'Size' cells
    
            for size, difficulties in pairs(sizes) do
                numdifficulties = 0 -- Span needed for the 'Size' cell above the 'Difficulty' cells

                if size > 5 or self.db.profile.show5Man then
                    for difficulty, _ in pairs(difficulties) do
                        colcount = colcount +1
                        numdifficulties = numdifficulties +1
        
                        lastcolumn = tooltip:AddColumn("CENTER")
                        tooltip:SetCell(3, lastcolumn, difficulty > 2 and self.db.profile.hc or self.db.profile.nhc)
    
                        raidorder[(string_format("%s.%d.%s", raid, size, difficulty))] = lastcolumn
                    end
                    tooltip:SetCell(2, lastcolumn - numdifficulties+1, size, numdifficulties)
                end
            end
            if colcount > 0 then
                tooltip:SetCell(1, lastcolumn - colcount+1, self.db.profile.instanceAbbr[raid], colcount)
            end
        end

        local currentPlayer, raids, currentRaid, lastline
        for currentPlayer, raids in pairs(charsdb) do
            lastline = tooltip:AddLine("")
            if self.db.profile.useClassColors then
                tooltip:SetCell(lastline, 1, currentPlayer, RAID_CLASS_COLORS_FONTS[self.db.global.charClass[currentPlayer]])
            else
                tooltip:SetCell(lastline, 1, currentPlayer)
            end

            for i = tooltip:GetColumnCount(),2,-1 do
                tooltip:SetCell(lastline, i, "")
                
                tooltip:SetCellColor(lastline, i, self.db.profile.freeraid.r, self.db.profile.freeraid.g, self.db.profile.freeraid.b, self.db.profile.freeraid.a)
            end
            for currentRaid, _ in pairs(raids) do
                if raidorder[currentRaid] then
                    tooltip:SetCellColor(lastline, raidorder[currentRaid], self.db.profile.savedraid.r, self.db.profile.savedraid.g, self.db.profile.savedraid.b, self.db.profile.savedraid.a)
                end
            end
        end
    else
        -- No saved raids at all
        tooltip:AddHeader(L["No saved raids found"])
    end
end


function Ailo:GetNextPurge()
    local charsdb = self.db.global.chars
    local currentPlayer, raids, currentRaid, expireTime
    local nextPurge = 0
    for currentPlayer, raids in pairs(charsdb) do 
        for currentRaid, expireTime in pairs(raids) do
            if nextPurge == 0 or nextPurge > expireTime then
                nextPurge = expireTime
            end
        end
    end
    return nextPurge
end

function Ailo:PurgeOldRaidIDs()
    local charsdb = self.db.global.chars
    local currentPlayer, raids, currentRaid, expireTime
    local now = time()
    for currentPlayer, raids in pairs(charsdb) do 
        for currentRaid, expireTime in pairs(raids) do
            if now > expireTime then
                self.db.global.chars[currentPlayer][currentRaid] = nil
            end
        end
        if not next(charsdb[currentPlayer]) then 
            self.db.global.chars[currentPlayer] = nil
        end
    end
end

function Ailo:ExtendRaidTable(instanceName, size, difficulty)
    -- Save it for the tooltip
    self.db.global.raids[instanceName] = self.db.global.raids[instanceName] or {}
    self.db.global.raids[instanceName][size] =  self.db.global.raids[instanceName][size] or {}
    self.db.global.raids[instanceName][size][difficulty] = true
    
    -- Add it to the abbreviations table
    -- If it has no abbreviation yet, try it with a somewhat good guess
    if not self.db.profile.instanceAbbr[instanceName] then
        self.db.profile.instanceAbbr[instanceName] = self:AbbreviateInstanceName(instanceName)
    end
end

function Ailo:AbbreviateInstanceName(instanceName)
    return string_gsub(instanceName, "(%a)[%l%p]*[%s%-]*", "%1")
end


function Ailo:TrimRaidTable()
    local charsdb = self.db.global.chars
    local currentPlayer, raids, currentRaid
    local raidsdb = self.db.global.raids
    local raid, size, difficulties,  difficulty
    local isUsed, raidString = false
    for raid, sizes in pairs(raidsdb) do
        for size, difficulties in pairs(sizes) do
            for difficulty, _ in pairs(difficulties) do
                isUsed = false
                raidString = string_format("%s.%d.%s", raid, size, difficulty)
                for currentPlayer, raids in pairs(charsdb) do 
                    if raids and raids[raidString] then
                        isUsed = true
                        break
                    end
                end
                if not isUsed then
                    self:DeleteFromRaidTable(raid, size, difficulty)
                end
            end
        end
    end
end

function Ailo:DeleteFromRaidTable(instanceName, size, difficulty)
    self.db.global.raids[instanceName][size][difficulty] = nil
    if not next(self.db.global.raids[instanceName][size]) then
        self.db.global.raids[instanceName][size] = nil
    end
    if not next(self.db.global.raids[instanceName]) then
        self.db.global.raids[instanceName] = nil
    end
end

function Ailo:ManualPlayerUpdate()
    local currentRaid
    if not self.db.global.chars[currentChar] then 
        self.db.global.chars[currentChar] = {}
    else 
        for currentRaid, _ in pairs(self.db.global.chars[currentChar]) do
            self.db.global.chars[currentChar][currentRaid] = nil
        end
    end
    self:Output(L["Updating data for current player."])
    self:PurgeOldRaidIDs()
    self:TrimRaidTable()
    self:UpdatePlayer()
end

function Ailo:SaveRaidForChar(char, instance, expireTime)
    if not self.db.global.chars[char] then
        self.db.global.chars[char] = {}
    end
    self.db.global.chars[char][instance] = expireTime
    
    if expireTime < self.db.global.nextPurge or self.db.global.nextPurge == 0 then
        self.db.global.nextPurge = expireTime
    end
end

function Ailo:WipeDB()
    if type(self.db.global.chars) == "table" then 
        wipe(self.db.global.chars)
    end
    if type(self.db.global.raids) == "table" then 
        wipe(self.db.global.raids)
    end
end

function Ailo:UPDATE_INSTANCE_INFO()
    self:UpdatePlayer()
end

function Ailo:UpdatePlayer()
    if 80 > UnitLevel("player") then return end
    self.db.global.charClass[currentChar] = select(2,UnitClass('player'))
    local now, index = time()
    local instanceReset, instanceDifficulty, locked, isRaid, maxPlayers
    local instanceNameAbbr, instanceString
    for index=1,GetNumSavedInstances() do
        instanceName, _, instanceReset, instanceDifficulty, locked, _, _, isRaid, maxPlayers, _ = GetSavedInstanceInfo(index)
        if locked then
            instanceString   = string_format("%s.%d.%s", instanceName, maxPlayers, instanceDifficulty)
            self:ExtendRaidTable(instanceName, maxPlayers, instanceDifficulty)
            self:SaveRaidForChar(currentChar, instanceString, now+instanceReset)
        end
    end
end


Ailo.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Ailo")
function Ailo:UPDATE_INSTANCE_INFO()
    self:UpdatePlayer()
end

local INSTANCE_SAVED = _G["INSTANCE_SAVED"]
function Ailo:CHAT_MSG_SYSTEM(event, msg)
    -- You are now saved to this instances.
    -- Refresh RaidInfo
    if tostring(msg) == INSTANCE_SAVED then
        RequestRaidInfo()
    end
end